"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.recordMatchesFilter = exports.recordMatchesRule = exports.getRecordValue = exports.getRuleValues = exports.matchers = exports.parseNumber = exports.ruleFunctionMap = void 0;
require("./polyfills.js");
exports.ruleFunctionMap = {
    '': 'ruleEquals',
    'not': 'ruleDoesNotEqual',
    'min': 'ruleMinimum',
    'max': 'ruleMaximum',
    'contains': 'ruleContains',
    'not-contains': 'ruleDoesNotContain',
    'starts-with': 'ruleStartsWith',
    'ends-with': 'ruleEndsWith',
    'in': 'ruleIn',
    'not-in': 'ruleNotIn'
};
exports.parseNumber = function (input) {
    return String(input)
        .replace('$', '')
        .replace(',', '');
};
exports.matchers = {
    ruleEquals: function (recordValue, ruleValue) {
        if (Array.isArray(recordValue)) {
            return recordValue.length === 0;
        }
        return String(ruleValue).toLowerCase() === String(recordValue).toLowerCase();
    },
    ruleDoesNotEqual: function (recordValue, ruleValue) {
        if (Array.isArray(recordValue)) {
            return recordValue.length > 0;
        }
        return String(ruleValue).toLowerCase() !== String(recordValue).toLowerCase();
    },
    ruleMinimum: function (recordValue, ruleValue) {
        if (Array.isArray(recordValue)) {
            return false;
        }
        if (recordValue === '') {
            return false;
        }
        if (ruleValue && !isNaN(Number(ruleValue))) {
            recordValue = exports.parseNumber(recordValue);
            try {
                ruleValue = parseFloat(ruleValue);
                recordValue = parseFloat(recordValue);
            }
            catch (err) {
                return false;
            }
            return recordValue >= ruleValue;
        }
        return recordValue >= ruleValue;
    },
    ruleMaximum: function (recordValue, ruleValue) {
        if (Array.isArray(recordValue)) {
            return false;
        }
        if (recordValue === '') {
            return false;
        }
        if (ruleValue && !isNaN(Number(ruleValue))) {
            recordValue = exports.parseNumber(recordValue);
            try {
                ruleValue = parseFloat(ruleValue);
                recordValue = parseFloat(recordValue);
            }
            catch (err) {
                return false;
            }
            return recordValue <= ruleValue;
        }
        return recordValue <= ruleValue;
    },
    ruleContains: function (recordValue, ruleValue) {
        if (Array.isArray(recordValue)) {
            if (!Array.isArray(ruleValue)) {
                ruleValue = [ruleValue];
            }
            for (var index = 0; index < ruleValue.length; index++) {
                if (recordValue.indexOf(ruleValue[index]) === -1) {
                    return false;
                }
            }
            return true;
        }
        return String(recordValue).indexOf(String(ruleValue)) !== -1;
    },
    ruleDoesNotContain: function (recordValue, ruleValue) {
        return !exports.matchers.ruleContains(recordValue, ruleValue);
    },
    ruleStartsWith: function (recordValue, ruleValue) {
        if (Array.isArray(recordValue)) {
            return false;
        }
        return String(recordValue).startsWith(String(ruleValue));
    },
    ruleEndsWith: function (recordValue, ruleValue) {
        if (Array.isArray(recordValue)) {
            return false;
        }
        return String(recordValue).endsWith(String(ruleValue));
    },
    ruleIn: function (recordValue, ruleValue) {
        if (!Array.isArray(ruleValue)) {
            return false;
        }
        if (Array.isArray(recordValue)) {
            for (var i = 0; i < recordValue.length; i++) {
                if (exports.matchers.ruleIn(recordValue[i], ruleValue)) {
                    return true;
                }
            }
            return false;
        }
        else {
            var iRecordValue = String(recordValue).toLowerCase();
            var iRuleValues = ruleValue.map(function (value) {
                return String(value).toLowerCase();
            });
            return iRuleValues.indexOf(iRecordValue) !== -1;
        }
    },
    ruleNotIn: function (recordValue, ruleValue) {
        if (!Array.isArray(ruleValue)) {
            return false;
        }
        return !exports.matchers.ruleIn(recordValue, ruleValue);
    }
};
exports.getRuleValues = function (filter) {
    if (typeof filter.value === 'string' && filter.value.indexOf('|') !== -1) {
        return filter.value.split('|');
    }
    if (filter.value === 'null' || filter.value === null) {
        return [''];
    }
    return [filter.value];
};
exports.getRecordValue = function (record, filter) {
    var attributePieces = String(filter.attribute).split('.');
    var recordValue = record;
    attributePieces.forEach(function (attributePiece) {
        recordValue = recordValue && recordValue[attributePiece];
    });
    if (recordValue) {
        if (recordValue.value !== undefined) {
            recordValue = recordValue.value;
        }
        else if (recordValue.id !== undefined) {
            recordValue = recordValue.id;
        }
    }
    if (recordValue === null || recordValue === undefined) {
        return '';
    }
    return recordValue;
};
exports.recordMatchesRule = function (record, rule, options) {
    if (options === void 0) { options = {}; }
    var operators = ['and', 'or'];
    if (operators.indexOf(Object.keys(rule)[0]) !== -1) {
        return exports.recordMatchesFilter(record, rule, options);
    }
    if (rule.filter !== undefined) {
        if (options.subfiltering) {
            var subRecord = record[rule.attribute];
            return exports.recordMatchesFilter(subRecord, rule.filter, options);
        }
        else {
            throw new Error('Subfilter matching is not supported');
        }
    }
    if (typeof rule.value === 'string' && rule.value.split('|').indexOf('logged-in-user') !== -1) {
        throw new Error('Dynamic filter conditions are not supported');
    }
    var recordValue = exports.getRecordValue(record, rule);
    var ruleValues = exports.getRuleValues(rule);
    var matchFunctionName = exports.ruleFunctionMap[rule.prefix];
    var matchFunction = exports.matchers[matchFunctionName];
    for (var i in ruleValues) {
        if (matchFunction(recordValue, ruleValues[i])) {
            return true;
        }
    }
    return false;
};
exports.recordMatchesFilter = function (record, filter, options) {
    if (options === void 0) { options = {}; }
    var currentOperator = Object.keys(filter)[0];
    if (filter[currentOperator].length === 0) {
        return true;
    }
    for (var i in filter[currentOperator]) {
        var match = exports.recordMatchesRule(record, filter[currentOperator][i], options);
        if (currentOperator === 'or' && match) {
            return true;
        }
        if (currentOperator === 'and' && !match) {
            return false;
        }
    }
    return currentOperator === 'and';
};
exports.default = exports.recordMatchesFilter;
